#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# License: GNU General Public License v2
#
# Author: thl-cmk[at]outlook[dot]com
# URL   : https://thl-cmk.hopto.org
# Date  : 2021-12-25
#
# rewritten for CMK 2.0 based on the check "Cisco Spanning-Tree Topology Change" (cisco_stptopo)
# by Udo Woehler (check-mk at bsw-com.de)
#
#
import time
from dataclasses import dataclass
from typing import Optional

from cmk.base.plugins.agent_based.agent_based_api.v1 import (
    register,
    Service,
    Result,
    State,
    SNMPTree,
    check_levels,
    not_startswith,
    get_value_store,
    Metric,
    OIDBytes,
)
from cmk.base.plugins.agent_based.agent_based_api.v1.type_defs import (
    DiscoveryResult,
    CheckResult,
    # StringTable,
    StringByteTable,
)


@dataclass
class StpTopo:
    last_change: int
    change_count: int
    root_mac: str
    root_cost: int

#
# [['527111100', '4', '\x80\x01\x00#\x04î¾\n']]
#
def parse_stp_topo_change(string_table: StringByteTable) -> Optional[StpTopo]:
    try:
        last_change, change_count, root_mac, root_cost = string_table[0]
    except IndexError:
        return

    if len(root_mac) == 8:  # vlan prefix
        root_mac = ':'.join(['%02s' % hex(m)[2:] for m in root_mac[2:]]).replace(' ', '0')
    elif len(root_mac) == 6:  # no prefix
        root_mac = ':'.join(['%02s' % hex(m)[2:] for m in root_mac]).replace(' ', '0')
    elif len(root_mac) == 22:  # Zyxel 4096/00:21:F7:DD:2C:00
        root_mac = ''.join(['%s' % chr(m) for m in root_mac[5:]]).replace(' ', '0')
    elif len(root_mac) == 23:  # Zyxel 4096/00:21:F7:DD:2C:00
        root_mac = ''.join(['%s' % chr(m) for m in root_mac[6:]]).replace(' ', '0')
    else:
        root_mac = 'N/A (wrong format)'

    return StpTopo(
        last_change=int(last_change) / 100 if last_change.isdigit() else 999999,
        change_count=int(change_count),
        root_mac=root_mac,
        root_cost=int(root_cost)
    )


def discovery_stp_topo_change(section: StpTopo) -> DiscoveryResult:
    yield Service()


def check_stp_topo_change(params, section: StpTopo) -> CheckResult:
    now = int(time.time())
    days, rest = divmod(section.last_change, 60 * 60 * 24)
    hours, rest = divmod(rest, 60 * 60)
    minutes, seconds = divmod(rest, 60)

    since = time.strftime('%c', time.localtime(now - section.last_change))

    value_store = get_value_store()

    last_topo1 = value_store.get('cisco.stp.topo', 1000000000)
    last_topo2 = value_store.get('cisco.stp.topo2', 1000000000)
    last_topo3 = value_store.get('cisco.stp.topo3', 1000000000)
    last_count = value_store.get('cisco.stp.count', 0)

    if last_count < section.change_count:  # stp change occurred, rotate counters
        value_store['cisco.stp.topo3'] = last_topo2
        value_store['cisco.stp.topo2'] = last_topo1
        value_store['cisco.stp.topo'] = section.last_change
        value_store['cisco.stp.count'] = section.change_count
    elif last_count > section.change_count:  # switch restart or counter overflow, reinit
        value_store['cisco.stp.topo3'] = 1000000000
        value_store['cisco.stp.topo2'] = 1000000000
        value_store['cisco.stp.topo'] = section.last_change
        value_store['cisco.stp.count'] = section.change_count

    warn, crit = params['levels_lower']

    yield Result(state=State.OK, summary=f'Root MAC {section.root_mac.upper()}')

    if section.root_cost == 0:
        yield Result(state=State.OK, summary=f'This bridge is the root')
    else:
        yield from check_levels(
            label='Cost to root bridge',
            value=section.root_cost,
            render_func=lambda v: str(v)
        )

    yield from check_levels(
        label='Changes',
        value=section.change_count,
        metric_name='stp_change_count',
        render_func=lambda v: str(v)
    )

    yield Result(state=State.OK, summary=f'Last change {since} ({days:.0f} {hours:02.0f}:{minutes:02.0f}:{seconds:02.0f})')

    if section.last_change <= warn and last_topo2 <= warn and last_topo3 <= warn and params['always_ok']:
        yield Result(
            state=State.OK,  
            summary=f'OK - Forced - Three times below threshold.'
        )
    elif section.last_change <= crit and last_topo2 <= crit and last_topo3 <= crit:
        yield Result(
            state=State.CRIT,
            summary=f'Three times below threshold.'
        )
    elif section.last_change <= warn and last_topo2 <= warn and last_topo3 <= warn:
        yield Result(
            state=State.WARN,
            summary=f'Three times below threshold.'
        )

    yield Metric(
        name='stp_last_topo',
        value=section.last_change,
        levels=(warn, crit),
    )


register.snmp_section(
    name='stp_topo_change',
    parse_function=parse_stp_topo_change,
    fetch=SNMPTree(
        base='.1.3.6.1.2.1.17.2',  # BRIDGE-MIB::dot1dStp
        oids=[
            '3',  # dot1dStpTimeSinceTopologyChange
            '4',  # dot1dStpTopChanges
            OIDBytes('5'),  # dot1dStpDesignatedRoot
            '6',  # dot1dStpRootCost
        ]
    ),
    detect=not_startswith('.1.3.6.1.2.1.17.2.4.0', '0'),
)

register.check_plugin(
    name='stp_topo_change',
    service_name='STP topology change',
    discovery_function=discovery_stp_topo_change,
    check_function=check_stp_topo_change,
    check_default_parameters={
        'always_ok': False,
        'levels_lower': (900, 600),
    },
    check_ruleset_name='stp_topo_change',
)
